/*
 * Title:        bits/message.hpp
 * Description:  C++ header-only wrapper for C-ABI data_t type
 * $Copyright (c) 2015-2017, 2020, 2022 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 */

/* This file:
 * Contains the C++ implementation of Message
 *
 */

#ifndef _DATAT_BITS_INCLUDE_TAG
#error Must only be included from sag_connectivity_cpp.hpp, use that instead
#endif

/**
 * A container for an payload and associated metadata.
 *
 * This class is not thread-safe, so users are required to ensure that
 * each instance of Message is only accessed by the codec or transport
 * plug-in that currently owns it, and that only one thread is
 * accessing it at any one time.
 */
class Message: protected sag_underlying_message_t
{
public:
	/** The type of a message metadata.
	 * @deprecated Use map_t instead
	 */
	typedef com::softwareag::_DATAT_INTERNAL_CPP_NAMESPACE::metadata_t metadata_t;
	/** The type of a message metadata. */
	typedef com::softwareag::_DATAT_INTERNAL_CPP_NAMESPACE::map_t map_t;
	/** The type of a message payload. */
	typedef data_t payload_t;
	/** Construct an empty message. */
	Message();

	/** Construct a message from a payload. This is a move constructor so the
		the argument will be left in an empty state. If you do not want this
		.copy() otherwise use std::move() */
	explicit Message(payload_t &&_payload);

	/** Construct a message from a payload and metadata. This is a move
		constructor so the the argument will be left in an empty state. If you
		do not want this use .copy() otherwise use std::move().
		@deprecated Use the map_t overload instead.
		*/
	Message(payload_t &&_payload, metadata_t &&_metadata);
	/** Construct a message from a payload and map_t. This is a move
		constructor so the the argument will be left in an empty state. If you
		do not want this use .copy() otherwise use std::move() */
	Message(payload_t &&_payload, map_t &&_metadata);

	/** Free the underlying payload and metadata */
	~Message();

	/** Move constructor from another message. The argument will be left in an
		empty state. If you do not want this use .copy() otherwise use
		std::move(). */
	Message(Message &&other);

	/** Move assignment from another message. The argument will be left in an
		empty state and our old contents freed. If you do not want this use
		.copy() otherwise use std::move(). */
	Message &operator=(Message &&other)
	{
		Message tmp;
		swap(std::move(tmp));
		swap(std::move(other));
		return *this;
	}
	/** Return a reference to the payload */
	const payload_t &getPayload() const;
	/** @deprecated
	 * Return a reference to the metadata 
	 */
	const metadata_t &getMetadata() const { return reinterpret_cast<const metadata_t&>(metadata); }

	/** Return a reference to the payload */
	payload_t &getPayload() { return static_cast<payload_t&>(payload); }
	/** @deprecated
	 * Return a reference to the metadata 
	 */
	metadata_t &getMetadata() { return reinterpret_cast<metadata_t&>(metadata); }
	/** Return a view on the metadata which can contain non-string values */
	map_t &getMetadataMap() { return static_cast<map_t&>(metadata); }
	/** Return a view on the metadata which can contain non-string values */
	const map_t &getMetadataMap() const { return static_cast<const map_t&>(metadata); }

	/** Set the payload. This is a move operation so the argument will be 
		left in an empty state and the old payload freed If you do not want this
		.copy() otherwise use std::move() */
	void setPayload(payload_t &&_payload);
	/** Set the metadata as metadata_t. This is a move operation so the argument will be 
		left in an empty state and the old metadata freed If you do not want this
		.copy() otherwise use std::move().
		@deprecated Use the map_t overload instead.
		*/
	void setMetadata(metadata_t &&_metadata);
	/** Set the metadata as map_t. This is a move operation so the argument will be 
		left in an empty state and the old metadata freed If you do not want this
		.copy() otherwise use std::move() */
	void setMetadata(map_t &&_metadata);

	/** Swap the contents of this message with another */
	void swap(Message &&other);

	/** Return a deep copy of this message, payload and metadata */
	Message copy() const
	{
		return Message(getPayload().copy(), getMetadata().copy());
	}

	/** Returns the metadata key used by the host for identifying the type of the message - "sag.type" */
	static const char *HOST_MESSAGE_TYPE() {
		return "sag.type";
	}

	/** Returns the metadata key used by the host for identifying the channel of the message - "sag.channel" */
	static const char *CHANNEL() {
		return "sag.channel";
	}

	/**
	 * The metadata key indicating that a message is a control message, and
	 * what its type is. The metadata value names the control message, with
	 * different types of control message defined in the CONTROL_TYPE_*
	 * constants.
	 *
	 * Control messages must have a null payload.
	 *
	 * @see CONTROL_TYPE_ACK_REQUIRED()
	 * @see CONTROL_TYPE_ACK_UPTO()
	 * @see CONTROL_TYPE_FLUSH()
	 * @see CONTROL_TYPE_FLUSH_ACK()
	 */
	static const char *CONTROL_TYPE() {
		return "sag.control.type";
	}

	/**
	 * The metadata key used for unique identification of towards-host
	 * messages. This is not mandatory, and only has to be populated for
	 * messages that you would like to see reliably acknowledged.
	 *
	 * The identifier can be an arbitrary value of string type, and must
	 * uniquely identify a message within the scope of the chain and the
	 * lifetime of the system. That is, possibly across host reboots and
	 * reconnections. Where applicable you should use an id provided by the
	 * external messaging system. This will help traceability of messages.
	 *
	 * @see CONTROL_TYPE_ACK_REQUIRED()
	 * @see CONTROL_TYPE_ACK_UPTO()
	 * @see CONTROL_TYPE()
	 */
	static const char* MESSAGE_ID() {
		return "sag.messageId";
	}

	/**
	 * The metadata key used for unique identification of towards-transport
	 * control messages, with a value of type int64_t. Used to correlate Flush
	 * and FlushAck control messages.
	 *
	 * @see CONTROL_TYPE_FLUSH()
	 * @see CONTROL_TYPE_FLUSH_ACK()
	 * @see CONTROL_TYPE()
	 */
	static const char* REQUEST_ID() {
		return "sag.requestId";
	}

	/**
	 * Name of the Flush control message, a towards-transport control message.
	 * It is a request for the transport to acknowledge that all previous
	 * towards-transport non-control messages have been fully and reliably
	 * processed. Metadata of this control message will also contain a unique
	 * REQUEST_ID.
	 *
	 * @see CONTROL_TYPE_FLUSH_ACK()
	 * @see REQUEST_ID()
	 * @see CONTROL_TYPE()
	 */
	static const char* CONTROL_TYPE_FLUSH() {
		return "Flush";
	}

	/**
	 * Name of the FlushAck control message, a towards-host control message in
	 * response to Flush, containing the corresponding REQUEST_ID. FlushAck
	 * messages implicitly cover all preceding Flush messages. That is, it is
	 * safe to miss a Flush if you have a later Flush in hand that succeeds it.
	 *
	 * @see CONTROL_TYPE_FLUSH()
	 * @see REQUEST_ID()
	 * @see CONTROL_TYPE()
	 */
	static const char* CONTROL_TYPE_FLUSH_ACK() {
		return "FlushAck";
	}

	/**
	 * Name of the AckRequired control message, a towards-host control message.
	 * It is a request for the host to acknowledge that all previous
	 * towards-host non-control messages have been fully and reliably
	 * processed. Metadata of this control message should also contain the
	 * MESSAGE_ID of the immediately preceding non-control message.
	 *
	 * @see CONTROL_TYPE_ACK_UPTO()
	 * @see MESSAGE_ID()
	 * @see CONTROL_TYPE()
	 */
	static const char* CONTROL_TYPE_ACK_REQUIRED() {
		return "AckRequired";
	}

	/**
	 * Name of the AckUpTo control message, a towards-transport control message
	 * in response to AckRequired, containing the corresponding MESSAGE_ID.
	 * AckUpTo messages will be seen in the exact order that their
	 * corresponding AckRequired messages were sent, with none missing or
	 * duplicated.
	 *
	 * @see CONTROL_TYPE_ACK_REQUIRED()
	 * @see MESSAGE_ID()
	 * @see CONTROL_TYPE()
	 */
	static const char* CONTROL_TYPE_ACK_UPTO() {
		return "AckUpTo";
	}

	/** Sets a value in the metadata. Despite taking a data_t this must only take string value. 
	 * Using this method may be more efficient even if you are copying the data_t.

		Note this overwrites a previous value if it exists, unlike the insert method on metadata_t. */
	Message &putMetadataValue(data_t &&key, data_t &&value);

private:
	/** This is not implicitly copiable. Use the copy() method. */
	Message(const Message &other) { abort(); }
	/** This is not implicitly copiable. Use the copy() method. */
	Message &operator=(const Message &other) { abort(); }
};


