/*
 * Title:        bits/metadata_t.hpp
 * Description:  C++ header-only wrapper for C-ABI data_t type
 * $Copyright (c) 2015-2017, 2019-2020, 2022 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 */

/* This file:
 * Contains the C++ implementation of metadata_t
 *
 */

#ifndef _DATAT_BITS_INCLUDE_TAG
#error Must only be included from sag_connectivity_cpp.hpp, use that instead
#endif

/**
 * A map class which implements many of the functions on std::map.
 *
 * This is an unordered map.
 *
 * We don't guarantee exact method-for-method compatibility with any of the
 * std types.
 *
 * This class gives access methods to a map of const char*->const char*.
 *
 * Keys starting with "sag." are reserved for internal use.
 *
 * This is moveable, but not implicitly copyable. If you want to copy it use
 * the .copy() method. If you want to move it use std::move().
 *
 * No underlying implementation of the map is guaranteed. Modifying the
 * contents will invalidate any iterators you have and any references that you
 * have to the underlying structure.
 *
 * @deprecated Use map_t instead.
 */
class metadata_t: protected map_t
{
	friend class data_t;
	friend class Message;
public:
	/**
	 * Forward/reverse and const/non-const iterators are implemented using this
	 * class.
	 *
	 * It implements the std::bidirectional_iterator concept: http://en.cppreference.com/w/cpp/concept/BidirectionalIterator
	 *
	 * You must not modify the metadata object while iterating over it. The iterator will throw (or assert on a non-NDEBUG build)
	 * if you try deferencing it after modifying the metadata object.
	 */
	template<typename DATA, typename UNDERLYING, typename PAIR>
	struct _iterator: public map_t::_iterator<DATA, UNDERLYING, PAIR>
	{
		typedef _iterator<DATA, UNDERLYING, PAIR> iterator;
		typedef map_t::_iterator<DATA, UNDERLYING, PAIR> map_iterator;

		_iterator(): map_iterator() {}
		_iterator(const metadata_t *meta, int64_t offs, bool stepToItem=false)
			: map_iterator(meta, offs, stepToItem)
		{}
		_iterator(const map_iterator &other): map_iterator(other) {}
		iterator &operator=(const map_iterator &other)
		{
			map_iterator::tabledata = other.tabledata;
			map_iterator::offs = other.offs;
			map_iterator::step = other.step;
			map_iterator::capacity = other.capacity;
			return *this;
		}

		std::string key() const 
		{
			assert(map_iterator::tabledata[map_iterator::offs].key.tag == SAG_DATA_STRING || map_iterator::tabledata[map_iterator::offs].key.tag == SAG_DATA_CONST_STRING);
			if (map_iterator::tabledata[map_iterator::offs].key.tag != SAG_DATA_STRING && map_iterator::tabledata[map_iterator::offs].key.tag != SAG_DATA_CONST_STRING) throw std::runtime_error("Metadata key not a string");
			return map_iterator::tabledata[map_iterator::offs].key.string;
		}
		std::string value() const
		{
		   try {
		      return(convert_to<std::string>(map_iterator::operator*().second));
		   } catch (...) {
		      throw std::runtime_error("Non-stringifiable data in metadata");
		   }
		}
	private: // can't expose these as string/string so make them inaccessible
		typename map_iterator::element_t &operator*() const
		{
			return reinterpret_cast<typename map_iterator::element_t &>(map_iterator::tabledata[map_iterator::offs]);
		}
		typename map_iterator::element_t *operator->() const
		{
			return reinterpret_cast<typename map_iterator::element_t *>(map_iterator::tabledata + map_iterator::offs);
		}
	};
	typedef _iterator<data_t, sag_underlying_map_table_entry_t *, pair<data_t> > iterator;
	typedef _iterator<const data_t, sag_underlying_map_table_entry_t const *, const pair<const data_t> > const_iterator;

	/** Construct an empty map */
	metadata_t(): map_t() {}
	/** Construct a map with (at least) capacity for n items */
	explicit metadata_t(size_t n): map_t(n) {}
	/** Move constructor. Other will be left as the empty map */
	metadata_t(metadata_t &&other) : map_t(std::move(other)) {}
	/** Move assignment. Other will be left as the empty map and our previous
		contents will be freed */
	metadata_t &operator=(metadata_t &&other)
	{
		metadata_t tmp;
		swap(std::move(tmp));
		swap(std::move(other));
		return *this;
	}
	/** Return a deep copy of this map */
	metadata_t copy() const;

	using map_t::empty;
	using map_t::size;
	using map_t::clear;

	/** Returns true if this map deep-equals other */
	bool operator==(const metadata_t &other) const
	{
		return static_cast<const map_t&>(*this) == static_cast<const map_t&>(other);
	}
	/** Returns false if this map deep-equals other */
	bool operator!=(const metadata_t &other) const { return !operator==(other); }

	/** Swap the contents of this map and other. */
	void swap(metadata_t &&other)
	{
		return static_cast<map_t&>(*this).swap(static_cast<map_t&&>(other));
	}

	/** Returns the item with the given key.
		If k is not in the map then throws std::runtime_error
		This does not return a reference which can be assigned to */
	std::string operator[](const char *k)
	{
		iterator it = find(k);
		if (it == end()) {
			throw std::runtime_error("Element not in map");
		} else {
			return it.value();
		}
	}
	/** Returns the item with the given key.
		If k is not in the map then throws std::runtime_error */
	std::string const operator[](const char *k) const
	{
		const_iterator it = find(k);
		if (it == end()) {
			throw std::runtime_error("Element not in map");
		} else {
			return it.value();
		}
	}
	/** Returns the item with the given key.
		If k is not in the map then throws std::runtime_error
		This does not return a reference which can be assigned to */
	std::string operator[](const std::string &k)
	{
		iterator it = find(k);
		if (it == end()) {
			throw std::runtime_error("Element not in map");
		} else {
			return it.value();
		}
	}
	/** Returns the item with the given key.
		If k is not in the map then throws std::runtime_error */
	std::string const operator[](const std::string &k) const
	{
		const_iterator it = find(k);
		if (it == end()) {
			throw std::runtime_error("Element not in map");
		} else {
			return it.value();
		}
	}
	/** Searches the map for an item with the given key and returns an iterator
		to that position in the map. If it is not found will return end(). */
	iterator find(const char *k)
	{
		return map_t::find(k);
	}
	/** Searches the map for an item with the given key and returns an iterator
		to that position in the map. If it is not found will return end(). */
	const_iterator find(const char *k) const
	{
		return map_t::find(k);
	}
	/** Searches the map for an item with the given key and returns an iterator
		to that position in the map. If it is not found will return end(). */
	iterator find(const std::string &k)
	{
		return map_t::find(k);
	}
	/** Searches the map for an item with the given key and returns an iterator
		to that position in the map. If it is not found will return end(). */
	const_iterator find(const std::string &k) const
	{
		return map_t::find(k);
	}

	/** Searches the map for an item with the given key and returns an iterator
		to that position in the map. If it is not found will return end(). */
	iterator find(const data_t &k)
	{
		return map_t::find(k);
	}
	/** Searches the map for an item with the given key and returns an iterator
		to that position in the map. If it is not found will return end(). */
	const_iterator find(const data_t &k) const
	{
		return map_t::find(k);
	}
	/**
	 * Insert a new key/value pair into the map. If the key already exists
	 * then the map is not modified and an iterator to the existing item is
	 * returned.
	 * Returns pair<iterator, bool> where the iterator points to the item with
	 * the given key and the boolean is true if the map needed to be modified
	 * with the new pair.
	 */
	std::pair<iterator, bool> insert(const std::pair<std::string, std::string> &v)
	{
		return map_t::insert(std::make_pair(data_t(v.first), data_t(v.second)));
	}

	/**
	 * Insert a new key/value pair into the map. If the key already exists
	 * then the map is not modified and an iterator to the existing item is
	 * returned.
	 * Returns pair<iterator, bool> where the iterator points to the item with
	 * the given key and the boolean is true if the map needed to be modified
	 * with the new pair.
	 */
	std::pair<iterator, bool> insert(const std::pair<const char*, const char*> &v)
	{
		return map_t::insert(std::make_pair(data_t(v.first), data_t(v.second)));
	}

	/**
	 * Remove the item with the specified key. Returns the number of items
	 * removed (always 1)
	 */
	size_t erase(const std::string &k)
	{
		return map_t::erase(k);
	}
	/**
	 * Remove the item with the specified key. Returns the number of items
	 * removed (always 1)
	 */
	size_t erase(const char *k)
	{
		return map_t::erase(k);
	}
	/**
	 * Erase the item pointed to by this iterator. If the iterator is end()
	 * this is a noop.
	 */
	void erase(iterator it)
	{
		map_t::erase(it);
	}

	/* Iteration: 
	 * begin() and end() methods, both const and non-const. Unordered so no reverse iterator
	 */
	/** Forward iterator begin */
	iterator begin() { return iterator(this, 0, true); }
	/** Forward iterator end */
	iterator end() { return iterator(this, table?table->capacity:0, false); }
	/** Forward const_iterator begin */
	const_iterator begin() const { return const_iterator(this, 0, true); }
	/** Forward const_iterator end */
	const_iterator end() const { return const_iterator(this, table?table->capacity:0, false); }
	/** Forward const_iterator begin */
	const_iterator cbegin() const { return const_iterator(this, 0, true); }
	/** Forward const_iterator end */
	const_iterator cend() const { return const_iterator(this, table?table->capacity:0, false); }

private:
	/** This is not implicitly copiable. Use the copy() method. */
	metadata_t &operator=(const metadata_t &other) { abort(); }
	/** This is not implicitly copiable. Use the copy() method. */
	metadata_t(const metadata_t &other) { abort(); }
};


