/*
 * engine_client_cpp.hpp
 *
 * $Copyright(c) 2002-2012 Progress Software Corporation and/or its subsidiaries and affiliates. All rights reserved.$
 * $Copyright (c) 2013-2019, 2021-2022 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 *
 */

/** @file
 * This is the header file for the Apama C++ EngineManagement API. 
 */

#ifndef ENGINE_CLIENT_CPP_HPP
#define ENGINE_CLIENT_CPP_HPP

#include <AP_Types.h>
#include <AP_Platform.h>
#include <iostream>
#include <sstream>
#include <exception>
#include <iterator>
#include <algorithm>
#include <utility>
#include <string>
#include <assert.h>
#include <sag_internal/engine_client_c.h>
#include <client_status_names.hpp>
#include <iomanip>
#include <vector>


using std::pair;

#undef ERROR

namespace com {

/** Contains classes for writing client applications for Apama components. */
namespace apama {

	/**
	 * An EngineException (or instance of a derived class) is thrown by 
	 * methods in this library if any problems are encountered.
	 */
	class AP_ENGINE_CLIENT_API EngineException : public std::exception
	{
	public:
		/** Construct an EngineException from a message */
		EngineException(const char *message);
		/** Copy constructor for EngineException */
		EngineException(const EngineException& ce);

		/** Copy assignment operator */
		EngineException& operator= (const EngineException& ce) {
			EngineException(ce).swap(*this);
			return *this;
		}

		/** Destructor */
		virtual ~EngineException() throw();

		/**
		 * Retrieve the message enclosed within the exception.
		 *
		 * @return C style string with details of the error that
		 * caused the operation to fail.
		 */
		virtual const char* what() const throw() { return m_message; }

		/**
		 * Retrieve the set of warnings associated with the exception
		 *
		 * @return A pointer to a NULL terminated list of warnings,
		 * possibly NULL if there are none. This list is owned by
		 * the exception object. It should not be deleted.
		 */
		virtual const char* const* getWarnings() const { return m_warnings; }

	protected:
		/** \internal swap contents with another exception */
		void swap(EngineException &);
		/** \internal the list of warnings associated with this exception */
		const char* const* m_warnings;

	private:
		/** \internal the message of the exception */
		char *m_message;
	};


	/** Available logging levels */
	typedef enum {
		/** Nothing logged */
		OFF_LEVEL,
		/** Startup and other important messages */
		FORCE_LEVEL,
		/** Critical messages */
		CRIT_LEVEL,
		/** Fatal messages */
		FATAL_LEVEL,
		/** Error messages */
		ERROR_LEVEL,
		/** Warning messages */
		WARN_LEVEL,
		/** Information messages (default) */
		INFO_LEVEL,
		/** Debugging messages */
		DEBUG_LEVEL,
		/** More-verbose debug messages */
		TRACE_LEVEL
	} LogLevel;


	/**
	 * Sets the level at which the client library will log information.
	 *
	 * @param level The level to log at
	 */
	AP_ENGINE_CLIENT_API void setLogLevel(LogLevel level);

	/**
	 * Sets the file to which the client library should log information.
	 *
	 * @param filename The filename to which the library should log. Will log
	 * to stderr if the filename cannot be opened.
	 * @param truncate If non-zero the log file will be truncated when it is
	 * opened. If zero then it will simply be appended to.
	 * @param utf8 Interpret filename as UTF-8, rather than the local character
	 * set.
	 */
	AP_ENGINE_CLIENT_API void setLogFile(const AP_char8* filename, bool truncate, bool utf8 = false);

	/**
	* Get the file descriptor of this logger.
	*
	*/
	AP_ENGINE_CLIENT_API int getLogFD();
			
	/** @internal
	 * Sets the mutex which the client library logger should use. For internal use only.
	 *
	 */
	AP_ENGINE_CLIENT_API void setLogMutex(void *mutex);
		
	/**
	 * Re-opens the log file. Also called if SIGHUP is received.
	 *
	 */
	AP_ENGINE_CLIENT_API void reOpenLog();
		
		
	/** Convert a string in local encoding to UTF-8, as required by most of the Apama API */
	AP_ENGINE_CLIENT_API AP_char8* convertToUTF8(const AP_char8* s);

	/** Convert a string in UTF-8 to the local encoding, since most of the Apama API returns UTF-8 */
	AP_ENGINE_CLIENT_API AP_char8* convertFromUTF8(const AP_char8* s);

	/**
	 * Free any char* string returned from the client API
	 */
	AP_ENGINE_CLIENT_API void char8free(char *string);

	/**
	 * Contains classes for handling events passed over the client API
	 */
	namespace event {

		/**
		 * An Event object represents an event instance.
		 *
		 * Must be created using com::apama::event::createEvent
		 * Must be destroyed using com::apama::event::deleteEvent
		 *
		 * Event is not copiable
		 */
		class AP_ENGINE_CLIENT_API Event
		{
			friend AP_ENGINE_CLIENT_API Event* createEvent(const char* eventString, const char* channelString);
			friend AP_ENGINE_CLIENT_API void deleteEvent(Event* ev);

		public:
			/**
			 * Retrieve the event's type and its contents as a string.
			 *
			 * @return C style string with the event's textual representation.
			 * Note that these strings should be encoded in UTF-8.
			 */
			virtual const char* getText() const = 0;

			/**
			 * Retrieve the event's channel.  This currently only has meaning for
			 * events which have been sourced from the correlator, in which case
			 * it is set to the name of the channel the event was emitted from, or
			 * "" (the empty string) for the wildcard channel.
			 *
			 * @return The channel of the event, encoded in UTF-8.
			 */
			virtual const char* getChannel() const = 0;

			/**
			 * Retrieve the event's timestamp.  This timestamp currently only has
			 * meaning for events which have been sourced from the correlator, in
			 * which case it is set to the correlator time at which it was created,
			 * or the value it was set to explicitly by code running within the
			 * correlator.
			 *
			 * @return the timestamp of the event, in floating point seconds since
			 * the Unix epoch.
			 */
			 virtual double getTime() const = 0;

			/** Output to an ostream from a reference */
			inline friend std::ostream& operator << (std::ostream& stream, const Event& obj) {
				stream << obj.getText();
				return stream;
			}
			/** Output to an ostream from a pointer */
			inline friend std::ostream& operator << (std::ostream& stream, const Event* obj) {
				stream << obj->getText();
				return stream;
			}

		private:
			/** \internal not publically copiable */
			Event(const Event&);
			/** \internal not publically copiable */
			Event& operator= (const Event&);

		protected:
			/** \internal Should be constructed using createEvent */
			Event();
			/** \internal Should be destroyed using deleteEvent */
			virtual ~Event();
		};


		/**
		 * An EventSupplier represents the resources created by the Engine
		 * to service a connection to an external sink of events. It
		 * filters the event output of the Engine by delivering only the
		 * events emitted on a particular set of channels. An
		 * EventSupplier passes events to an EventConsumer.
		 * EventSupplier objects should be freed using the
		 * com::apama::event::deleteEventSupplier function.
		 *
		 * EventSupplier is not copiable
		 */
		class AP_ENGINE_CLIENT_API EventSupplier
		{
			friend AP_ENGINE_CLIENT_API void deleteEventSupplier(EventSupplier* evsup);

		public:
			/**
			 * Disconnect the EventSupplier from its consumer.
			 *
			 * Disconnection is automatically performed as part of
			 * deleteEventSupplier so in most cases there is no need to call
			 * this method explicitly.
			 *
			 * @exception EngineException
			 */
			virtual void disconnect() = 0;

		protected:
			/** \internal can't be directly constructed */
			EventSupplier();
			/** \internal must be freed with deleteEventSupplier */
			virtual ~EventSupplier();

		private:
			/** \internal not copiable */
			EventSupplier(const EventSupplier&);
			/** \internal not copiable */
			EventSupplier& operator= (const EventSupplier&);
		};


		/**
		 * An EventConsumer can connect to the Engine through an
		 * EventSupplier and register to receive events. In order to
		 * receive events from the Engine, a developer must inherit from
		 * this class and define its sendEvents method. This method is
		 * called by the EventSupplier when events are emitted from the
		 * Engine.
		 *
		 * EventConsumers are not copiable
		 */
		class AP_ENGINE_CLIENT_API EventConsumer
		{
		public:
			/**
			 * This method must be defined in inherited classes to
			 * enable receiving of events. This method is called by an
			 * EventSupplier.
			 *
			 * Note that EventSuppliers are not reentrant so
			 * calling the disconnect method on the calling
			 * EventSupplier is not permitted from within the
			 * sendEvents implementation.
			 *
			 * @param events A null-terminated array of pointers to Event objects.
			 */
			virtual void sendEvents(const Event* const* events) = 0;

		protected:
			/** \internal Not directly constructable (obviously, it's pure virtual) */
			EventConsumer();
			/** \internal Not directly deletable (really we should be giving you a reference to make this clear) */
			virtual ~EventConsumer();

		private:
			/** \internal not copiable */
			EventConsumer(const EventConsumer&);
			/** \internal not copiable */
			EventConsumer& operator= (const EventConsumer&);
		};

		/**
		 * Derived class from EventConsumer which implements the disconnect API.
		 */
		class  AP_ENGINE_CLIENT_API DisconnectableEventConsumer : public EventConsumer
		{
		public:
			/**
			 * Used to inform the consumer that it is not going
			 * to be sent any more events.
			 *
			 * @param reason A string giving the reason why the consumer
			 * is being disconnected. May be NULL.
			 */
			virtual void disconnect(const char* reason) = 0;
			
		protected:
			/** \internal Not directly constructable (obviously, it's pure virtual) */
			DisconnectableEventConsumer();
			/** \internal Not directly deletable (really we should be giving you a reference to make this clear) */
			virtual ~DisconnectableEventConsumer();
			
		private:
			/** \internal not copiable */
			DisconnectableEventConsumer(const DisconnectableEventConsumer&);
			/** \internal not copiable */
			DisconnectableEventConsumer& operator= (const DisconnectableEventConsumer&);
		};


		/**
		 * This function allows creation of an Event object.
		 *
		 * @param eventString C style string representing the event
		 * instance in EPL. Note that this string should
		 * be encoded in UTF-8.
		 * @param channelString C style string representing the
		 * channel
		 * @return A reference to an Event object.
		 */
		AP_ENGINE_CLIENT_API Event* createEvent(const char* eventString, const char* channelString);

		/**
		 * This function allows deletion of an Event object.
		 *
		 * @param ev A reference to an Event object.
		 */
		AP_ENGINE_CLIENT_API void deleteEvent(Event* ev);


		/**
		 * This function disconnects and deletes an EventSupplier object.
		 *
		 * It also stops the associated EventConsumer from receiving
		 * events.
		 *
		 * @param evsup	A reference to an EventSupplier object.
		 */
		AP_ENGINE_CLIENT_API void deleteEventSupplier(EventSupplier* evsup);

	} // namespace event

	/** Contains classes for interacting with the correlator component */
	namespace engine {

		/**
		 * A MonitorScript object encapsulates an EPL code
		 * fragment, containing package, event and monitor definitions to
		 * be injected into an Engine.
		 *
		 * Must be created using com::apama::engine::createMonitorScript.
		 * Must be delete using com::apama::engine::deleteMonitorScript.
		 */
		class AP_ENGINE_CLIENT_API MonitorScript
		{
			friend AP_ENGINE_CLIENT_API MonitorScript* createMonitorScript(const char* monitorString);
			friend AP_ENGINE_CLIENT_API void deleteMonitorScript(MonitorScript* mon);

		public:
			/**
			 * Retrieve the text of a MonitorScript fragment as a string.
			 *
			 * @return C style string containing the text of the
			 * EPL fragment. Note that this string should
			 * be encoded in UTF-8.
			 */
			virtual const char* getText() const = 0;

			/** Output a reference to a std::ostream  */
			inline friend std::ostream& operator << (std::ostream& stream, const MonitorScript& obj) {
				stream << obj.getText();
				return stream;
			}
			/** Output a pointer to a std::ostream */
			inline friend std::ostream& operator << (std::ostream& stream, const MonitorScript* obj) {
				stream << obj->getText();
				return stream;
			}

		private:
			/** \internal not copiable */
			MonitorScript(const MonitorScript&);
			/** \internal not copiable */
			MonitorScript& operator= (const MonitorScript&);

		protected:
			/** \internal Not directly constructable (use createMonitorScript) */
			MonitorScript();
			/** \internal Not directly deleteable (use deleteMonitorScript) */
			virtual ~MonitorScript();
		};

		class EngineStatus;
		std::ostream& operator << (std::ostream& stream, const EngineStatus& obj);
		std::ostream& operator << (std::ostream& stream, const EngineStatus* obj);


		/**
		 * EngineStatus represents the operational status of the Engine.
		 *
		 * EngineStatus objects are returned by com::apama::engine::EngineManagement::getStatus
		 */
		class AP_ENGINE_CLIENT_API EngineStatus
		{
			friend class EngineManagement; /* getStatus */

		public:

			/**
			 * A class used for the iterating through all status items.
			 *
			 * Implements the ForwardIterator concept: http://en.cppreference.com/w/cpp/concept/ForwardIterator
			 */
#ifdef _MSC_VER
// needed or MSVC complains about the use of pair<string> here, even though it's all inlined
#pragma warning (disable : 4251)
#endif
			class AP_ENGINE_CLIENT_API StatusIterator : public std::iterator<std::forward_iterator_tag, int>
			{
			public:
				/** element type */
				typedef pair<std::string, std::string> pair_t;

				/** \internal this should be instantiated via EngineStatus::begin() */
				explicit StatusIterator(const EngineStatus &es, bool end=false);
				/** Default-constructed status iterator. Use EngineStatus::begin() to get a valid iterator */
				StatusIterator():
					m_status(0),
					m_opaqueState(0),
					iterPair()
				{}
				/** Destroy underling status iterator */
				~StatusIterator()
				{
					if (m_status && m_opaqueState) m_status->disposeIterator(m_opaqueState);
				}
				/** (postfix) increment this iterator */
				StatusIterator operator++(int dummy/*@unused@*/)
				{
					StatusIterator current(*this);
					++(*this); //this->operator++() // prefix ++
					return current;
				}
				/** (prefix) increment this iterator */
				StatusIterator& operator++();
				/** Returns true if two iterators point to the same item */
				bool operator==(const StatusIterator& ref) const
				{
					return (m_status == ref.m_status) &&
							(((0 == m_opaqueState) && (0 == ref.m_opaqueState)) ||
							(iterPair.first == ref.iterPair.first));
				}
				/** Returns false if two iterators point to the same item */
				bool operator!=(const StatusIterator& ref) const { return !((*this)==ref); }
				/** Return a reference to the element pointed to */
				pair_t &operator*() { return iterPair; }
				/** Return a pointer to the element pointed to */
				pair_t *operator->() { return &iterPair; }
				/** Returns true if this is a valid iterator over status data */
				bool valid() { return 0 != m_opaqueState; }

				/** Copy constructor */
				StatusIterator(const StatusIterator& ref)
					: m_status(ref.m_status),
					m_opaqueState(m_status->copyIterator(ref.m_opaqueState)),
					iterPair(ref.iterPair)
				{}
				/** Copy assignment */
				StatusIterator &operator=(const StatusIterator &other)
				{
					StatusIterator(other).swap(*this);
					return *this;
				}

			private:
				/** \internal step this iterator */
				void step();
				/** \internal swap the contents of this iterator with another */
				void swap(StatusIterator &other)
				{
					std::swap(m_status, other.m_status);
					std::swap(m_opaqueState, other.m_opaqueState);
					std::swap(iterPair, other.iterPair);
				}

				/** \internal pointer to the EngineStatus this iterator was from */
				const EngineStatus *m_status;
				/** \internal pointer to the underling iterator state */
				void *m_opaqueState;
				/** \internal the data item that's currently being pointed to */
				pair_t iterPair;
			};

			/** Iterator type */
			typedef StatusIterator iterator;
			/** Iterator type for const EngineStatus */
			typedef StatusIterator const_iterator;

			/**
			 * The time in milliseconds since the correlator was started.
			 */
			AP_uint64 getUptime() const {
				return getStatusIntValue(client_status::UPTIME_SECONDS);
			}

			/**
			* The number of contexts in the correlator, including the main context.
			*/
			AP_uint32 getNumContexts() const {
				return (AP_uint32) getStatusIntValue(client_status::CONTEXTS_TOTAL);
			}

			/**
			* The number of EPL monitor definitions injected into the correlator. 
			*/
			AP_uint32 getNumMonitors() const {
				return (AP_uint32) getStatusIntValue(client_status::MONITORS_TOTAL);
			}

			/**
			* The number of monitor instances, also known as sub-monitors.
			*/
			AP_uint32 getNumProcesses() const {
				return (AP_uint32) getStatusIntValue(client_status::MONITOR_INSTANCES_TOTAL);
			}

			/**
			* The number of Java applications and Java EPL plug-ins loaded in the correlator. 
			*/
			AP_uint32 getNumJavaApplications() const {
				return (AP_uint32) getStatusIntValue(client_status::JAVA_APPLICATIONS_TOTAL);
			}

			/**
			* The number of listeners in all contexts. 
			*/
			AP_uint32 getNumListeners() const {
				return (AP_uint32) getStatusIntValue(client_status::LISTENERS_TOTAL);
			}

			/**
			* The number of event types defined within the correlator. 
			*/
			AP_uint32 getNumEventTypes() const {
				return (AP_uint32) getStatusIntValue(client_status::EVENT_TYPES_TOTAL);
			}

			/**
			* The sum of routed events on the route queues of all contexts.
			*/
			AP_uint32 getNumQueuedFastTrack() const {
				return (AP_uint32) getStatusIntValue(client_status::QUEUED_ROUTE_TOTAL);
			}

			/**
			* The number of executors on the input queues of all contexts. 
			* As well as events, this can include clock ticks, spawns, injections and other operations. 
			*/
			AP_uint32 getNumQueuedInput() const {
				return (AP_uint32) getStatusIntValue(client_status::QUEUED_INPUT_TOTAL);
			}

			/**
			* The number of events that the correlator has received from external sources since the correlator started.
			*/
			AP_uint64 getNumReceived() const {
				return getStatusIntValue(client_status::INPUT_TOTAL);
			}

			/**
			* The number of events that have been routed across all contexts since the correlator was started.
			*/
			AP_uint64 getNumFastTracked() const {
				return getStatusIntValue(client_status::ROUTE_TOTAL);
			}

			/**
			* The number of external consumers/receivers connected to receive emitted events.
			*/
			AP_uint32 getNumConsumers() const {
				return (AP_uint32) getStatusIntValue(client_status::CONSUMERS_TOTAL);
			}

			/**
			* The number of events waiting on output queues to be dispatched to any connected external consumers/receivers.
			*/
			AP_uint32 getNumOutEventsQueued() const {
				return (AP_uint32) getStatusIntValue(client_status::QUEUED_OUTPUT_TOTAL);
			}

			/**
			* The number of events that have been sent or emitted to channels which 
			* have at least one external consumer/receiver subscribed.
			*/
			AP_uint64 getNumOutEventsCreated() const {
				return getStatusIntValue(client_status::CREATED_OUTPUT_TOTAL);
			}

			/**
			* The number of events that have been delivered to external consumers/receivers. 
			* This counts for each external consumer/receiver an event is sent to. 
			*/
			AP_uint64 getNumOutEventsSent() const {
				return getStatusIntValue(client_status::OUTPUT_TOTAL);
			}

			/**
			* The number of events on the input queues of all public contexts. 
			*/
			AP_uint32 getNumInputQueuedInput() const {
				return (AP_uint32) getStatusIntValue(client_status::QUEUED_INPUT_PUBLIC_TOTAL);
			}

			/**
			* The name of the slowest context.
			*/
			const AP_char8* getMostBackedUpInput() const {
				return (*this)[client_status::SLOWEST_INPUT_QUEUE_CONTEXT_NAME];
			}

			/**
			* The name of the consumer/receiver with the largest number of incoming events waiting to be processed.
			*/
			const AP_char8 *getSlowestReceiver() const {
				return (*this)[client_status::SLOWEST_OUTPUT_QUEUE_NAME];
			}

			/**
			* The number of events on the slowest consumer's/receiver's queue, as identified by the name of the slowest consumer/receiver.
			*/
			AP_uint32 getSlowestReceiverQueueSize() const {
				return (AP_uint32) getStatusIntValue(client_status::SLOWEST_OUTPUT_QUEUE_SIZE_TOTAL);
			}

			/**
			* The number of events on the slowest context's queue, as identified by the name of the slowest context.
			*/
			AP_uint32 getMostBackedUpQueueSize() const {
				return (AP_uint32) getStatusIntValue(client_status::SLOWEST_INPUT_QUEUE_SIZE_TOTAL);
			}

			/**
			* The number of sub-event-listeners that are active across all contexts.
			*/
			AP_uint32 getNumSubListeners() const {
				return (AP_uint32) getStatusIntValue(client_status::SUB_LISTENERS_TOTAL);
			}

			/**
			* The number of events processed by the correlator in all contexts.
			*/
			AP_uint64 getNumProcessed() const {
				return getStatusIntValue(client_status::PROCESSED_TOTAL);
			}
			/**
			 * Returns an instance of the StatusIterator which
			 * allows to iterate over all status items.
			 *
			 * @return StatusIterator object pointing to the first element
			 */					
			const_iterator begin() const {
				return StatusIterator(*this);
			}

			/**
			 * Returns an instance of the StatusIterator which
			 * allows to iterate over all status items.
			 *
			 * @return StatusIterator object pointing after the last element
			 */					
			const_iterator end() const {
				return StatusIterator(*this, true);
			}

			/**
			 * Returns the value of the status item with the given key
			 */
			const AP_char8* operator[] (const char *ref) const {
			    return getStatusValue(ref);
			}

			/**
			 * Returns the value of the status item with the given key
			 */
			const AP_char8* operator[] (std::string &ref) const {
				return this->operator[](ref.c_str());
			}

			/**
			* Get a user value of type AP_int64.
			*
			* @exception EngineException.
			* @return The user status.
			*/
			AP_int64 getUserInt(const char* key) {
				return userInt(key);
			}

			/**
			* Get a user value of type AP_int64 or the alt value if it cannot be found.
			*
			* @exception EngineException.
			* @return The user status.
			*/
			AP_int64 getUserIntOr(const char* key, int alt) {
				return userIntOr(key, alt);
			}

			/**
			* Get a user value of type AP_float64
			*
			* @exception EngineException
			* @return The user status.
			*/
			AP_float64 getUserFloat(const char* key) {
				return userFloat(key);
			}

			/**
			* Get a user value of type AP_float64 or the alt value if it cannot be found.
			*
			* @exception EngineException.
			* @return The user status.
			*/
			AP_float64 getUserFloatOr(const char* key, float alt) {
				return userFloatOr(key, alt);
			}

			/**
			* Get a user value of type AP_bool.
			*
			* @exception EngineException.
			* @return The user status.
			*/
			AP_bool getUserBool(const char* key) {
				return userBool(key);
			}

			/**
			* Get a user value of type AP_bool or the alt value if it cannot be found.
			*
			* @exception EngineException.
			* @return The user status.
			*/
			AP_bool getUserBoolOr(const char* key, bool alt) {
				return userBoolOr(key, alt);
			}

			/**
			* Get a user value of type AP_char8*.
			*
			* @exception EngineException.
			* @return The user status.
			*/
			const AP_char8* getUserString(const char* key) {
				return userString(key);
			}

			/**
			* Get a user value of type AP_char8* or the alt value if it cannot be found.
			*
			* @exception EngineException.
			* @return The user status.
			*/
			const AP_char8* getUserStringOr(const char* key, const char* alt) {
				return userStringOr(key, alt);
			}

			/**
			* Get all the status values available, formatted for terminal output. This includes any user statuses.
			*
			* @return The standard and user status values.
			*/
			std::string getAllStatusValues() {
				std::ostringstream ost;
				ost << getStandardStatusValues();
				for (EngineStatus::StatusIterator it = begin(); it != end(); ++it) {
					if (it->first.find("user-") != std::string::npos)
						ost << std::setw(32) << setiosflags(std::ios_base::left) << it->first.append(":") << it->second << std::endl;
				}
				return ost.str();
			}

			/**
			* Get the standard status values, formatted for terminal output.
			*
			* @return The standard status values.
			*/
			std::string getStandardStatusValues() {
				std::ostringstream ost;
				ost << *this;
				return ost.str();
			}

			/**
			* Get a header for the raw output, useful for indentifying values.
			*
			* @return The standard status header.
			*/	
			static std::string getRawHeader() {
				std::ostringstream ost;
				ost << "Uptime (ms),# Contexts,# Monitors,# Sub-monitors,# Java apps,# Listeners,";
				ost << "# Sub-listeners,# Event types,Input queue,# Received events,";
				ost << "Route queue,# Routed events,# Consumers,Output queue,";
				ost << "# Created output events,# Sent output events,# Processed events,";
				ost << "Slowest context name,Slowest context queue size,Slowest receiver,Slowest receiver queue";
				return ost.str();
			}

			/**
			* Get a header for the raw output, useful for indentifying values. This includes any user statuses.
			*
			* @return The standard and user status header.
			*/
			std::string getRawHeaderAll() {
				std::ostringstream ost;
				ost << getRawHeader();
				for (EngineStatus::StatusIterator it = begin(); it != end(); ++it) {
					if (it->first.find("user-") != std::string::npos)
						ost << "," << it->first;
				}
				return ost.str();
			}

			/**
			* Get the standard status values, formatted for parsers (values separated by ,).
			*
			* @return The standard status values.
			*/
			std::string getStandardStatusValuesRaw() {
				std::ostringstream ost;
				ost << getUptime()
				<< "," << getNumContexts()
				<< "," << getNumMonitors()
				<< "," << getNumProcesses()
				<< "," << getNumJavaApplications()
				<< "," << getNumListeners()
				<< "," << getNumSubListeners()
				<< "," << getNumEventTypes()
				<< "," << getNumQueuedInput()
				<< "," << getNumReceived()
				<< "," << getNumQueuedFastTrack()
				<< "," << getNumFastTracked()
				<< "," << getNumConsumers()
				<< "," << getNumOutEventsQueued()
				<< "," << getNumOutEventsCreated()
				<< "," << getNumOutEventsSent()
				<< "," << getStatusValue(client_status::PERSISTENCE_SNAPSHOTS_TOTAL)
				<< "," << getStatusValue(client_status::PERSISTENCE_LAST_SNAPSHOT_TIMESTAMP)
				<< "," << getStatusValue(client_status::PERSISTENCE_SNAPSHOT_WAIT_EWMA_SECONDS)
				<< "," << getStatusValue(client_status::PERSISTENCE_COMMIT_TIME_EWMA_SECONDS)
				<< "," << getStatusValue(client_status::PERSISTENCE_SNAPSHOT_ROWS_CHANGED_EWMA_TOTAL)
				<< "," << getNumProcessed()
				<< "," << getMostBackedUpInput()
				<< "," << getMostBackedUpQueueSize()
				<< "," << getSlowestReceiver()
				<< "," << getSlowestReceiverQueueSize();
				return ost.str();
			}

			/**
			* Get all the status values available, formatted for parsers (values separated by ,). This includes any user statuses.
			*
			* @return The standard and user status values.
			*/
			std::string getAllStatusValuesRaw() {
				std::ostringstream ost;
				ost << getStandardStatusValuesRaw();
				for (EngineStatus::StatusIterator it = begin(); it != end(); ++it) {
					if (it->first.find("user-") != std::string::npos)
						ost << "," << it->second;
				}
				return ost.str();
			}

			/**
			 * Destroy the underlying status data.
			 */
			~EngineStatus()
			{
				if (c_status) {
					c_status->functions->deleteStatus(reinterpret_cast<AP_EngineStatus *>(c_status));
				}
			}
			/** Copy constructor */
			EngineStatus(const EngineStatus &other)
				: c_status(other.c_status ? reinterpret_cast<AP_EngineStatus_API*>(other.c_status->functions->copyStatus(reinterpret_cast<AP_EngineStatus *>(other.c_status))) : 0)
			{}
			/** Copy assignment */
			EngineStatus &operator=(const EngineStatus &other)
			{
				EngineStatus(other).swap(*this);
				return *this;
			}
			/** Create a default-constructed (empty) EngineStatus.
			 * Valid EngineStatus objects are returned by EngineManagement::getStatus.
			 */
			EngineStatus(): c_status(0) {}
		private:
			/** \internal Should only be created by EngineManagement::getStatus */
			EngineStatus(AP_EngineStatus_API *api): c_status(api) {}
			/** \internal swap two EngineStatus objects */
			void swap(EngineStatus &other) {
				std::swap(c_status, other.c_status);
			}
			/** \internal only used by StatusIterator */
			void getNext(const char *&name, const char *&value, void *&opaqueState) const {
				if (c_status)
					c_status->functions->getNextStatusItem(reinterpret_cast<AP_EngineStatus *>(c_status), &name, &value, &opaqueState);
			}
			/** \internal Get a status value with a given key as a string */
			const AP_char8* getStatusValue(const char *name) const {
				return c_status->functions->getStatusValue(reinterpret_cast<AP_EngineStatus *>(c_status), name);
			}
			/** \internal Get a status value with a given key as an integer */
			AP_uint64 getStatusIntValue(const char *name) const {
				return c_status->functions->getStatusIntValue(reinterpret_cast<AP_EngineStatus *>(c_status), name);
			}
			/** \internal Used by StatusIterator to destroy underlying state */
			void disposeIterator(void *opaqueState) const {
				c_status->functions->disposeIterator(reinterpret_cast<AP_EngineStatus *>(c_status), opaqueState);
			}
			/** \internal Used by StatusIterator to copy underlying state */
			void *copyIterator(void *opaqueState) const {
				return c_status->functions->copyIterator(reinterpret_cast<AP_EngineStatus *>(c_status), opaqueState);
			}

			/** \internal Getter for user values */
			AP_int64 userIntOr(const char* key, AP_int64 alt) {
				return c_status->functions->getUserIntOr(reinterpret_cast<AP_EngineStatus *>(c_status), key, alt);
			}
			/** \internal Getter for user values */
			AP_float64 userFloatOr(const char* key, AP_float64 alt) {
				return c_status->functions->getUserFloatOr(reinterpret_cast<AP_EngineStatus *>(c_status), key, alt);
			}
			/** \internal Getter for user values */
			AP_bool userBoolOr(const char* key, AP_bool alt) {
				return c_status->functions->getUserBoolOr(reinterpret_cast<AP_EngineStatus *>(c_status), key, alt);
			}
			/** \internal Getter for user values */
			const AP_char8* userStringOr(const char* key, const AP_char8* alt) {
				return c_status->functions->getUserStringOr(reinterpret_cast<AP_EngineStatus *>(c_status), key, alt);
			}
			/** \internal Getter for user values */
			AP_int64 userInt(const char* key) {
				return c_status->functions->getUserInt(reinterpret_cast<AP_EngineStatus *>(c_status), key);
			}
			/** \internal Getter for user values */
			AP_float64 userFloat(const char* key) {
				return c_status->functions->getUserFloat(reinterpret_cast<AP_EngineStatus *>(c_status), key);
			}
			/** \internal Getter for user values */
			AP_bool userBool(const char* key) {
				return c_status->functions->getUserBool(reinterpret_cast<AP_EngineStatus *>(c_status), key);
			}
			/** \internal Getter for user values */
			const AP_char8* userString(const char* key) {
				return c_status->functions->getUserString(reinterpret_cast<AP_EngineStatus *>(c_status), key);
			}
			/** \internal Pointer to underlying C status object */
			AP_EngineStatus_API *c_status;
		};


		/**
		* Stream output operator for a reference.
		*/
		inline std::ostream& operator << (std::ostream& stream, const EngineStatus& obj) {
			std::ostringstream ost;
			ost
			<< "Uptime(ms):                                   " << obj.getUptime() << std::endl
			<< "Number of contexts:                           " << obj.getNumContexts() << std::endl
			<< "Number of monitors:                           " << obj.getNumMonitors() << std::endl
			<< "Number of sub-monitors:                       " << obj.getNumProcesses() << std::endl
			<< "Number of Java applications:                  " << obj.getNumJavaApplications() << std::endl
			<< "Number of listeners:                          " << obj.getNumListeners() << std::endl
			<< "Number of sub-listeners:                      " << obj.getNumSubListeners() << std::endl
			<< "Number of event types:                        " << obj.getNumEventTypes() << std::endl
			<< "Events on input queue:                        " << obj.getNumQueuedInput() << std::endl
			<< "Events received:                              " << obj.getNumReceived() << std::endl
			<< "Events processed:                             " << obj.getNumProcessed() << std::endl
			<< "Events on internal queue:                     " << obj.getNumQueuedFastTrack() << std::endl
			<< "Events routed internally:                     " << obj.getNumFastTracked() << std::endl
			<< "Number of consumers:                          " << obj.getNumConsumers() << std::endl
			<< "Events on output queue:                       " << obj.getNumOutEventsQueued() << std::endl
			<< "Output events created:                        " << obj.getNumOutEventsCreated() << std::endl
			<< "Output events sent:                           " << obj.getNumOutEventsSent() << std::endl
			<< "Events on input context queues:               " << obj.getNumInputQueuedInput() << std::endl
			<< "Number of persistent snapshots:               " << obj[client_status::PERSISTENCE_SNAPSHOTS_TOTAL] << std::endl
			<< "Last persistent snapshot time:                " << obj[client_status::PERSISTENCE_LAST_SNAPSHOT_TIMESTAMP] << std::endl
			<< "Persistent snapshot wait time Millis (EWMA):  " << obj[client_status::PERSISTENCE_SNAPSHOT_WAIT_EWMA_SECONDS] << std::endl
			<< "Persistent commit time Millis (EWMA):         " << obj[client_status::PERSISTENCE_COMMIT_TIME_EWMA_SECONDS] << std::endl
			<< "Last persistent snapshot rows changed (EWMA): " << obj[client_status::PERSISTENCE_SNAPSHOT_ROWS_CHANGED_EWMA_TOTAL] << std::endl
			<< "Slowest context name:                         " << obj.getMostBackedUpInput() << std::endl
			<< "Slowest context queue size:                   " << obj.getMostBackedUpQueueSize() << std::endl
			<< "Slowest receiver name:                        " << obj.getSlowestReceiver() << std::endl
			<< "Slowest receiver queue size:                  " << obj.getSlowestReceiverQueueSize() << std::endl;
			stream << ost.str();
			return stream;
		}

		/**
		* Stream output operator for a pointer.
		*/
		inline std::ostream& operator << (std::ostream& stream, const EngineStatus* obj) {
			stream << *obj;
			return stream;
		}

		inline EngineStatus::StatusIterator::StatusIterator(const EngineStatus &es, bool end):
			m_status(&es),
			m_opaqueState(0),
			iterPair()
		{
			if (!end) step(); // get the first item in the map
		}

		/** prefix ++ */
		inline EngineStatus::StatusIterator& EngineStatus::StatusIterator::operator++()
		{
			if (m_opaqueState) step();// don't step past the end
			return *this;
		}

		inline void EngineStatus::StatusIterator::step()
		{
			const char *name = nullptr;
			const char *value = nullptr;

			/* get the next item from the library map */
			m_status->getNext(name, value, m_opaqueState );
			if(!name || !value) {
				assert(!m_opaqueState);
			} else {
				iterPair = pair_t(name, value);
			}
		}

		/**
		 * Base class for a named object (i.e. event type, container type
		 * or monitor) returned by an engine inspection. Returned by methods
		 * of the EngineInfo class. Strings returned by methods of this class
		 * are valid until the EngineInfo class is deleted.
		 *
		 * NameInfo objects are not copiable.
		 * NameInfo objects are not directly constructable. Sub-classes are returned
		 * from functions on EngineInfo
		 */
		class AP_ENGINE_CLIENT_API NameInfo
		{
		public:
			/**
			 * Name, excluding package, for example "MyEvent".
			 */
			virtual const char* getName() const = 0;

			/**
			 * Package name, for example "com.apamax", or an
			 * empty string if in the  default package.
			 */
			virtual const char* getPackage() const = 0;

			/**
			 * Fully qualified name, for example "com.apamax.MyEvent"
			 */
			virtual const char* getFullyQualifiedName() const = 0;

		private:
			/** \internal not copiable */
			NameInfo(const NameInfo&);
			/** \internal not copiable */
			NameInfo& operator= (const NameInfo&);

		protected:
			/** \internal sub-classes are retuned from EngineInfo */
			NameInfo();
			/** \internal not deletable (we should be returning references) */
			virtual ~NameInfo();
		};

		/**
		 * Information about a monitor returned by an engine inspection.
		 * Returned by the getMonitors method of the EngineInfo class.
		 */
		class AP_ENGINE_CLIENT_API NamedMonitorInfo : public NameInfo
		{
		public:
			/**
			 * Gets the number of sub-monitors belonging to this monitor.
			 */
			virtual unsigned int getNumSubMonitors() const = 0;

		private:
			/** \internal not copiable */
			NamedMonitorInfo(const NamedMonitorInfo&);
			/** \internal not copiable */
			NamedMonitorInfo& operator= (const NamedMonitorInfo&);

		protected:
			/** \internal Returned from EngineInfo */
			NamedMonitorInfo();
			/** \internal not deleteable */
			virtual ~NamedMonitorInfo();
		};

		/**
		 * Information about a java application returned by an engine inspection.
		 * Returned by the getJavaApplications method of the EngineInfo class.
		 */
		class AP_ENGINE_CLIENT_API NamedJavaApplicationInfo : public NameInfo {

		public:
			/**
			 * Gets the number of listeners belonging to this application.
			 */
			virtual unsigned int getNumListeners() const = 0;

		private:
			/** \internal not copiable */
			NamedJavaApplicationInfo(const NamedJavaApplicationInfo&);
			/** \internal not copiable */
			NamedJavaApplicationInfo& operator= (const NamedJavaApplicationInfo&);

		protected:
			/** \internal Returned from EngineInfo */
			NamedJavaApplicationInfo();
			/** \internal not deleteable */
			virtual ~NamedJavaApplicationInfo();
		};
					
		/**
		 * Information about a context returned by an engine inspection.
		 * Returned by the getContexts method of the EngineInfo class.
		 */
		class AP_ENGINE_CLIENT_API NamedContextInfo : public NameInfo {

		public:
			/**
			 * Gets the number of sub-monitors belonging to this context.
			 */
			virtual unsigned int getNumSubMonitors() const = 0;

			/**
			 * Gets the queue size of this context.
			 */
			virtual unsigned int getQueueSize() const = 0;

			/**
			 * Gets the channels that this context is subscribed to (0-terminated array of 0-terminated strings)
			 */
			virtual char const *const *getChannels() const = 0;

		private:
			/** \internal not copiable */
			NamedContextInfo(const NamedContextInfo&);
			/** \internal not copiable */
			NamedContextInfo& operator= (const NamedContextInfo&);

		protected:
			/** \internal Returned from EngineInfo */
			NamedContextInfo();
			/** \internal not deleteable */
			virtual ~NamedContextInfo();
		};


		/**
		 * Information about a receiver returned by an engine inspection.
		 * Returned by the getReceivers method of the EngineInfo class.
		 */
		class AP_ENGINE_CLIENT_API NamedReceiverInfo : public NameInfo {

		public:
			/**
			 * Gets the queue size of this receiver
			 */
			virtual unsigned int getQueueSize() const = 0;

			/**
			 * Gets the channels that this receiver is subscribed to (0-terminated array of 0-terminated strings)
			 */
			virtual char const *const *getChannels() const = 0;

			/**
			 * Gets the address from which this receiver is connected
			 */
			virtual const AP_char8 *getAddress() const = 0;

		private:
			/** \internal not copiable */
			NamedReceiverInfo(const NamedReceiverInfo&);
			/** \internal not copiable */
			NamedReceiverInfo& operator= (const NamedReceiverInfo&);

		protected:
			/** \internal Returned from EngineInfo */
			NamedReceiverInfo();
			/** \internal not deleteable */
			virtual ~NamedReceiverInfo();
		};


		/**
		 * Information about a receiver returned by an engine inspection.
		 * Returned by the getReceivers method of the EngineInfo class.
		 */
		class AP_ENGINE_CLIENT_API NamedPluginReceiverInfo : public NameInfo {

		public:
			/**
			 * Gets the queue size of this receiver
			 */
			virtual unsigned int getQueueSize() const = 0;

			/**
			 * Gets the channels that this receiver is subscribed to (0-terminated array of 0-terminated strings)
			 */
			virtual char const *const *getChannels() const = 0;

		private:
			/** \internal not copiable */
			NamedPluginReceiverInfo(const NamedPluginReceiverInfo&);
			/** \internal not copiable */
			NamedPluginReceiverInfo& operator= (const NamedPluginReceiverInfo&);

		protected:
			/** \internal Returned from EngineInfo */
			NamedPluginReceiverInfo();
			/** \internal not deleteable */
			virtual ~NamedPluginReceiverInfo();
		};

		/**
		 * Information about a event type returned by an engine inspection.
		 * Returned by the getEventTypes method of the EngineInfo class.
		 */
		class AP_ENGINE_CLIENT_API NamedEventTypeInfo : public NameInfo {

		public:
			/**
			 * Gets the number of event templates for this event type.
			 */
			virtual unsigned int getNumEventTemplates() const = 0;

		private:
			/** \internal not copiable */
			NamedEventTypeInfo(const NamedEventTypeInfo&);
			/** \internal not copiable */
			NamedEventTypeInfo& operator= (const NamedEventTypeInfo&);

		protected:
			/** \internal Returned from EngineInfo */
			NamedEventTypeInfo();
			/** \internal not deleteable */
			virtual ~NamedEventTypeInfo();
		};

		/**
		 * Information about a timer type returned by an engine inspection.
		 * Returned by the getTimers method of the EngineInfo class.
		 */
		class AP_ENGINE_CLIENT_API NamedTimerInfo : public NameInfo {

		public:
			/**
			 * Gets the number of timers for this timer type.
			 */
			virtual unsigned int getNumTimers() const = 0;

		private:
			/** \internal not copiable */
			NamedTimerInfo(const NamedTimerInfo&);
			/** \internal not copiable */
			NamedTimerInfo& operator= (const NamedTimerInfo&);

		protected:
			/** \internal Returned from EngineInfo */
			NamedTimerInfo();
			/** \internal not deleteable */
			virtual ~NamedTimerInfo();
		};

		/**
		 * Information about an aggregate function returned by an engine
		 * inspection.
		 * Returned by the getAggregates method of the EngineInfo class.
		 */
		class AP_ENGINE_CLIENT_API NamedAggregateInfo : public NameInfo {

		private:
			/** \internal not copiable */
			NamedAggregateInfo(const NamedAggregateInfo&);
			/** \internal not copiable */
			NamedAggregateInfo& operator= (const NamedAggregateInfo&);

		protected:
			/** \internal Returned from EngineInfo */
			NamedAggregateInfo();
			/** \internal not deleteable */
			virtual ~NamedAggregateInfo();
		};
			
		/**
		 * Information about the monitors and types currently in an engine.
		 * Instances of this class are returned by EngineManagement::inspectEngine
		 * and can be deleted with the com::apama::engine::deleteInfo function.
		 *
		 * When an EngineInfo class is deleted, everything returned by the
		 * getMonitors, getEventTypes, etc methods is deleted.
		 * This includes the arrays themselves, the classes pointed to by
		 * the arrays and any strings returned by those classes. This means
		 * that deleteInfo is the only cleanup method that needs to be called
		 * after an engine inspection.
		 */
		class AP_ENGINE_CLIENT_API EngineInfo
		{
			friend AP_ENGINE_CLIENT_API void deleteInfo(EngineInfo* info);

		public:
			/**
			 * Gets the number of monitors in the engine.
			 */
			virtual unsigned int getNumMonitors() const = 0;

			/**
			 * Gets the number of Java applications in the engine.
			 */
			virtual unsigned int getNumJavaApplications() const = 0;

			/**
			 * Gets the number of event listener types in the engine.
			 */					
			virtual unsigned int getNumEventTypes() const = 0;	

			/**
			 * Gets the number of timers in the engine.
			 */					
			virtual unsigned int getNumTimers() const = 0;	

			/**
			 * Gets the number of aggregate functions in the engine.
			 */					
			virtual unsigned int getNumAggregates() const = 0;

			/**
			 * Returns information about the monitors in the engine,
			 * in the form of a NULL terminated array of pointers to
			 * MonitorInfo objects. The size of the array can be found
			 * by calling getNumMonitors (or looking for the
			 * NULL terminator).
			 */					
			virtual NamedMonitorInfo** getMonitors() const = 0;

			/**
			 * Returns information about the Java applications in the
			 * engine, in the form of a NULL terminated array of pointers
			 * to JavaApplicationInfo objects. The size of the array can
			 * be found by calling getNumJavaApplications (or looking for
			 * the NULL terminator).
			 */					
			virtual NamedJavaApplicationInfo** getJavaApplications() const = 0;

			/**
			 * Returns information about the event listener types in the engine,
			 * in the form of a NULL terminated array of pointers to
			 * NamedEventTypeInfo objects. The size of the array can be
			 * found by calling getNumEventTypes (or looking for the
			 * NULL terminator).
			 */					
			virtual NamedEventTypeInfo** getEventTypes() const = 0;

			/**
			 * Returns information about the timers in the engine,
			 * in the form of a NULL terminated array of pointers to
			 * NamedTimerInfo objects. The size of the array can be
			 * found by calling getNumTimers (or looking for the
			 * NULL terminator).
			 */					
			virtual NamedTimerInfo** getTimers() const = 0;

			/**
			 * Returns information about the aggregate functions in the
			 * engine, in the form of a NULL terminated array of pointers
			 * to NamedAggregateInfo objects. The size of the array can be
			 * found by calling getNumAggregates (or looking for the
			 * NULL terminator).
			 */					
			virtual NamedAggregateInfo** getAggregates() const = 0;

			/**
			 * Gets the number of contexts in the engine.
			 */
			virtual unsigned int getNumContexts() const = 0;

			/**
			 * Returns information about the contexts in the
			 * engine, in the form of a NULL terminated array of pointers
			 * to ContextInfo objects. The size of the array can
			 * be found by calling getNumContexts (or looking for
			 * the NULL terminator).
			 */					
			virtual NamedContextInfo** getContexts() const = 0;

			/**
			 * Gets the number of receivers connected to the engine
			 */
			virtual unsigned int getNumReceivers() const = 0;

			/**
			 * Returns information about the receivers connected to the engine
			 * in the form of a NULL-terminated array of pointers to
			 * NamedReceiverInfo objects. The size of the array can be
			 * can be found by calling getNumContexts or looking for the NULL
			 * terminator.
			 */
			virtual NamedReceiverInfo **getReceivers() const = 0;

			/**
			 * Gets the number of plugin receivers in the engine
			 */
			virtual unsigned int getNumPluginReceivers() const = 0;

			/**
			 * Returns information about the plugin receivers in the engine
			 * in the form of a NULL-terminated array of pointers to
			 * NamedReceiverInfo objects. The size of the array can be
			 * can be found by calling getNumContexts or looking for the NULL
			 * terminator.
			 */
			virtual NamedPluginReceiverInfo **getPluginReceivers() const = 0;

		private:
			/** \internal not copiable */
			EngineInfo(const EngineInfo&);
			/** \internal not copiable */
			EngineInfo& operator= (const EngineInfo&);

		protected:
			/** \internal Created by EngineManagement::inspectEngine */
			EngineInfo();
			/** \internal Destroyed by deleteInfo */
			virtual ~EngineInfo();
		};

		class AP_ENGINE_CLIENT_API EngineConnection;

		/**
		 * The Engine Management class acts as the interface to the Engine,
		 * and allows operations to be carried out on it. It is not
		 * possible to construct an object of this class. The static
		 * methods supplied below in the engine namespace must be used
		 * instead.
		 *
		 * Because an EngineManagement object is also an EventConsumer,
		 * it can have events sent to it (and to the Engine) through its
		 * sendEvents() method.
		 *
		 * EngineManagement objects are not copiable.
		 * EngineManagement objects are created using com::apama::engine::connectToEngine,
		 * com::apama::engine::connectToEngineReceiveOnly or com::apama::engine::connectToEngineMonitorOnly.
		 * EngineManagement objects are deleted using com::apama::engine::disconnectFromEngine.
		 */
		class AP_ENGINE_CLIENT_API EngineManagement : public com::apama::event::EventConsumer
		{
		public:
			/** The mode that attachAsEventConsumerTo and detachAsEventConsumerFrom operate in */
			enum ConnectMode {
				/** For backwards compatibility, single serial connection */
				CONNECT_LEGACY,
				/** Allow parallel connections on multiple channels */
				CONNECT_PARALLEL
			};
			/**
			 * Inject EPL text into the Engine.
			 *
			 * @param script The EPL text to be injected.
			 * @exception EngineException
			 */
			virtual void injectMonitorScript(MonitorScript& script) = 0;

			/**
			 * Inject EPL text into the Engine,
			 * returning any warnings produced by the
			 * correlator; maintained for compatibility only,
			 * please use injectMonitorScriptWithWarningsFilename instead of
			 * this function.
			 *
			 * @param script EPL text to be injected.
			 * @return A pointer to a NULL terminated array of
			 * warnings which must be deleted with the
			 * deleteWarnings() function, or NULL if there are
			 * none
			 * @exception EngineException, which may contain
			 * warnings as well as an error message
			 */
			virtual const char* const* injectMonitorScriptWithWarnings(MonitorScript& script) = 0;

			/**
			 * Inject EPL text into the Engine,
			 * supplying the filename it was injected from and
			 * returning any warnings produced by the
			 * correlator.
			 *
			 * @param script The EPL text to be injected.
			 * @param filename The filename the EPL text
			 * was injected from
			 * @return A pointer to a NULL terminated array of
			 * warnings which must be deleted with the
			 * deleteWarnings() function, or NULL if there are
			 * none
			 * @exception EngineException, which may contain
			 * warnings as well as an error message
			 */
			virtual const char* const* injectMonitorScriptWithWarningsFilename(MonitorScript& script, const char *filename) = 0;

			/**
			 * Delete a named object from the Engine.
			 *
			 * @param name	The name of the object to be deleted.
			 * @exception EngineException
			 */
			virtual void deleteName(const char * name) = 0;

			/**
			 * Force deletion of a named object from the Engine.
			 *
			 * @param name	The name of the object to be deleted.
			 * @exception EngineException
			 */
			virtual void forceDeleteName(const char * name) = 0;

			/**
			 * Kill a named monitor in the Engine.
			 *
			 * @param name	The name of the monitor to be killed.
			 * @exception EngineException
			 */
			virtual void killName(const char * name) = 0;

			/**
			 * Deletes everything from the engine.
			 */
			virtual void deleteAll() = 0;
					
			/**
			 * Injects a Java application (a jar) into the engine.
			 *
			 * @param jarbytes A pointer to the array of bytes
			 * containing the jar
			 * @param size The size of the jar
			 * @exception EngineException
			 */
			virtual void injectJava(const AP_uint8* jarbytes, AP_uint32 size) = 0;

			/**
			 * Injects a Java application (a jar) into the engine;
			 * maintained for compatibility only,
			 * please use injectJavaWithWarningsFilename instead of
			 * this function.
			 *
			 * @param jarbytes A pointer to the array of bytes
			 * containing the jar
			 * @param size The size of the jar
			 * @return A pointer to a NULL terminated array of
			 * warnings which must be deleted with the
			 * deleteWarnings() function, or NULL if there are
			 * none
			 * @exception EngineException, which may contain
			 * warnings as well as an error message
			 */
			virtual const char* const* injectJavaWithWarnings(const AP_uint8* jarbytes, AP_uint32 size) = 0;

			/**
			 * Injects a Java application (a jar) into the engine.
			 *
			 * @param jarbytes A pointer to the array of bytes
			 * containing the jar
			 * @param size The size of the jar
			 * @param filename The name of the jar file
			 * @return A pointer to a NULL terminated array of
			 * warnings which must be deleted with the
			 * deleteWarnings() function, or NULL if there are
			 * none
			 * @exception EngineException, which may contain
			 * warnings as well as an error message
			 */
			virtual const char* const* injectJavaWithWarningsFilename(const AP_uint8* jarbytes, AP_uint32 size, const char *filename) = 0;

					
			/**
			 * Injects a CDP (Correlator Deployment Package file) into the engine.
			 *
			 * @param cdpbytes A pointer to the array of bytes
			 * containing the CDP file
			 * @param size The size of the CDP
			 * @param filename The name of the CDP file
			 * @exception EngineException
			 */
			virtual void injectCDP(const AP_uint8* cdpbytes, AP_uint32 size, const char *filename=NULL) = 0;

			/**
			 * Injects a CDP (Correlator Deployment Package file) into the engine;
			 * maintained for compatibility only,
			 * please use injectCDPWithWarningsFilename instead of
			 * this function.
			 *
			 * @param cdpbytes A pointer to the array of bytes
			 * containing the CDP file
			 * @param size The size of the CDP file
			 * @return A pointer to a NULL terminated array of
			 * warnings which must be deleted with the
			 * deleteWarnings() function, or NULL if there are
			 * none
			 * @exception EngineException, which may contain
			 * warnings as well as an error message
			 */
			virtual const char* const* injectCDPWithWarnings(const AP_uint8* cdpbytes, AP_uint32 size) = 0;

			/**
			 * Injects a CDP (Correlator Deployment Package file) into the engine.
			 *
			 * @param cdpbytes A pointer to the array of bytes
			 * containing the CDP file
			 * @param size The size of the CDP file
			 * @param filename The name of the CDP file
			 * @return A pointer to a NULL terminated array of
			 * warnings which must be deleted with the
			 * deleteWarnings() function, or NULL if there are
			 * none
			 * @exception EngineException, which may contain
			 * warnings as well as an error message
			 */
			virtual const char* const* injectCDPWithWarningsFilename(const AP_uint8* cdpbytes, AP_uint32 size, const char *filename) = 0;

			/**
			 * Get the Engine's current operational status.
			 * @exception EngineException
			 */
			EngineStatus getStatus()
			{
				char *errmsg = nullptr;
				AP_EngineStatus *es_api = engineManagement_getStatusOpaque(this, &errmsg);
				if (errmsg) {
					com::apama::EngineException e(errmsg);
					free(errmsg);
					throw e;
				}

				return EngineStatus(reinterpret_cast<AP_EngineStatus_API *>(es_api));
			}

			/**
			 * Connect an event receiver to the Engine.
			 *
			 * @param consumer The EventConsumer to connect to the
			 * Engine.
			 * @param channels An array of names representing the
			 * channels to subscribe to. This is a null-terminated array
			 * of pointers to zero-terminated char arrays. If it is null or
			 * empty subscribe to all channels. Note that these channel
			 * names should be encoded in UTF-8.
			 * @param disconnectSlow tell the correlator it can disconnect
			 * this receiver if it is slow. Only the first consumer's
			 * disconnectSlow value is used; subsequent consumers added
			 * to this EngineManagement object share the connection and
			 * thus the disconnect behaviour. Defaults to false.
			 * @return A reference to an EventSupplier resource, which the caller
			 * is responsible for freeing, using com::apama::event::deleteEventSupplier().
			 * Each EventSupplier resource must be explicitly disconnected or deleted
			 * before disconnecting this EngineManagement object.
			 * @exception EngineException
			 * \deprecated This method is deprecated. Instead use EngineManagement::connectEventConsumer(com::apama::event::EventConsumer*, const char* const*)
			 * with com::apama::engine::connectToEngine(const char*, unsigned short, bool)
			 */
			virtual com::apama::event::EventSupplier* connectEventConsumer(com::apama::event::EventConsumer* consumer, const char* const* channels, bool disconnectSlow) = 0;

			/**
			* Connect an event receiver to the Engine.
			*
			* @param consumer The EventConsumer to connect to the
			* Engine.
			* @param channels An array of names representing the
			* channels to subscribe to. This is a null-terminated array
			* of pointers to zero-terminated char arrays. If it is null or
			* empty subscribe to all channels. Note that these channel
			* names should be encoded in UTF-8.
			* @return A reference to an EventSupplier resource, which the caller
			* is responsible for freeing, using com::apama::event::deleteEventSupplier().
			* Each EventSupplier resource must be explicitly disconnected or deleted
			* before disconnecting this EngineManagement object.
			* @exception EngineException
			*/
			virtual com::apama::event::EventSupplier* connectEventConsumer(com::apama::event::EventConsumer* consumer, const char* const* channels) = 0;

			/**
			 * Connect this Engine as an event receiver to another Engine.
			 *
			 * @param target The Engine to connect to
			 * @param channels An array of names representing the
			 * channels to subscribe to. This is a null-terminated array
			 * of pointers to zero-terminated char arrays. If this is null or empty,
			 * subscribe to all channels. Note that these channel names
			 * should be encoded in UTF-8.
			 * @param disconnectSlow disconnect if slow. Only the first consumer's
			 * disconnectSlow value is used; subsequent consumers added to this
			 * EngineManagement object share the connection and
			 * thus the disconnect behaviour. Default is false.
			 * @param mode the connection mode to use,
			 * defaults to legacy (single connection, all
			 * events delivered to the default
			 * channel). Set to CONNECT_PARALLEL for
			 * connection per channel and channel values
			 * passed through.
			 * @return true if successful
			 * @exception EngineException
			 */
			virtual bool attachAsEventConsumerTo(EngineManagement* target, const char* const* channels, bool disconnectSlow = false, ConnectMode mode = CONNECT_LEGACY) = 0;

					
			/**
			 * Connect this Engine as an event receiver to another Engine.
			 *
			 * @param host The hostname of the Engine to connect to
			 * @param port The port of the Engine to connect to
			 * @param channels An array of names representing the
			 * channels to subscribe to. This is a null-terminated array
			 * of pointers to zero-terminated char arrays. If this is null or empty,
			 * subscribe to all channels. Note that these channel names
			 * should be encoded in UTF-8.
			 * @param disconnectSlow disconnect if slow. Only the first consumer's
			 * disconnectSlow value is used; subsequent consumers added to this
			 * EngineManagement object share the connection and thus the disconnect behaviour.
			 * @param mode the connection mode to use,
			 * defaults to legacy (single connection, all
			 * events delivered to the default
			 * channel). Set to CONNECT_PARALLEL for
			 * connection per channel and channel values
			 * passed through.
			 * @return true if successful
			 * @exception EngineException
			 */
			virtual bool attachAsEventConsumerTo(const char* host, int port, const char* const* channels, bool disconnectSlow = false, ConnectMode mode = CONNECT_LEGACY) = 0;

					
			/**
			 * Unsubscribe as an event receiver from another engine.
			 *
			 * @param target The Engine to unsubscribe from.
			 * @param channels An array of names representing the
			 * channels to unsubscribe from. This is a null-terminated array
			 * of pointers to zero-terminated char arrays. If this is null or empty
			 * unsubscribe from all channels.
			 * @param mode the connection mode to use,
			 * defaults to legacy (single connection, all
			 * events delivered to the default
			 * channel). Set to CONNECT_PARALLEL for
			 * connection per channel and channel values
			 * passed through.
			 * @exception EngineException
			 */					
			virtual void detachAsEventConsumerFrom(EngineManagement* target, const char* const* channels, ConnectMode mode = CONNECT_LEGACY) = 0;
			
			/**
			 * Unsubscribe as an event receiver from another engine.
			 *
			 * @param host The host of the Engine to unsubscribe from.
			 * @param port The port of the Engine to unsubscribe from.
			 * @param channels An array of names representing the
			 * channels to unsubscribe from. This is a null-terminated array
			 * of pointers to zero-terminated char arrays. If this is null or empty
			 * unsubscribe from all channels.
			 * @param mode the connection mode to use,
			 * defaults to legacy (single connection, all
			 * events delivered to the default
			 * channel). Set to CONNECT_PARALLEL for
			 * connection per channel and channel values
			 * passed through.
			 * @exception EngineException
			 */					
			virtual void detachAsEventConsumerFrom(const char* host, int port, const char* const* channels, ConnectMode mode = CONNECT_LEGACY) = 0;

			/**
			 * Inject events into the Engine (inherited from
			 * EventConsumer), automatically batching messages
			 * on a separate thread. May return before events are
			 * sent - call flushEvents before exiting.
			 *
			 * @param events An array of pointers to Event objects
			 * containing the events to inject into the Engine.
			 * @exception EngineException
			 */
			virtual void sendEvents(const com::apama::event::Event* const* events) = 0;

			/**
			 * Inject events into the Engine
			 *
			 * @param events An array of pointers to Event objects
			 * containing the events to inject into the Engine.
			 * @exception EngineException
			 */
			virtual void sendEventsNoBatching(const com::apama::event::Event* const* events) = 0;

			/**
			 * Send any outstanding events sent via sendEvents.
			 */
			virtual void flushEvents() = 0;
			
			/**
			 * Returns information about the monitors, event types and
			 * container types which exist in the engine. Use deleteInfo
			 * to delete the returned object when it is no longer needed.
			 *
			 * @return Information about the engine.
			 * @exception EngineException
			 */
			virtual EngineInfo* inspectEngine() = 0;

			/**
			 * This method is used to check that the Engine is
			 * still alive. If the Engine is alive it returns
			 * normally. If there is a problem then an
			 * EngineException is thrown.
			 *
			 * At present, ping() will not attempt to
			 * reconnect, and will therefore always throw
			 * an exception once the connection is lost.
			 * A new management object should be created
			 * and the old one cleaned up (disconnected)
			 * when reconnection is needed.
			 *
			 *
			 * @exception EngineException
			 */
			virtual void ping() = 0;
					
			/**
			 * This method is used to check that this object is
			 * still connected to the Engine. It will never
			 * reconnect. Returns true if connected.
			 */
			virtual bool isConnected() = 0;

			/**
			Attaches this component as a persistent receiver of the remote
			components specified in the connections argument, then decrements the
			reference count of any existing connections (which typically results in
			their removal).
			
			@param connections A vector of EngineConnection objects. The method call doesn't modify the vector.
			
			This method throws EngineException If the change in the persistent connection set
			failed. Note that no exception is thrown if some of the specified
			connections could not be established immediately - in this case the
			receiver will simply keep trying to reconnect in the background.
			
			@exception EngineException
			*/
			virtual void setReceiverConnections(const std::vector<com::apama::engine::EngineConnection>& connections) = 0;

			/**
			Get all the persistent connections associated with the engine.
			@return an array of EngineConnection objects.
			*/
			virtual const std::vector<com::apama::engine::EngineConnection> getReceiverConnections() = 0;
					
		protected:
			/** \internal EngineManagement objects are created using connectToEngine, connectToEngineReceiveOnly or connectToEngineMonitorOnly. */
			EngineManagement();
			/** \internal EngineManagement objects are deleted using disconnectFromEngine. */
			virtual ~EngineManagement();

		private:
			/** \internal not copiable */
			EngineManagement(const EngineManagement&);
			/** \internal not copiable */
			EngineManagement& operator= (const EngineManagement&);
		};

		/**
		EngineConnection encapsulates the information related to a connection
		between two Apama messaging components.
		*/
		class AP_ENGINE_CLIENT_API EngineConnection {
		private:
			std::string host;
			int port;
			std::vector<std::string> channels;
			bool disconnectSlow;
			EngineManagement::ConnectMode connectMode;

			void init(const char * host, int port, bool disconnectSlow, EngineManagement::ConnectMode connectMode, const char * const * channels) {
				if (host == NULL || channels == NULL) {
					throw EngineException("EngineConnection cannot be created with null host or channels");
				}
				if (port < 0 || port > 65535) {
					throw EngineException("EngineConnection cannot be created with invalid port value");
				}

				this->host = std::string(host);
				this->port = port;
				this->disconnectSlow = disconnectSlow;
				this->connectMode = connectMode;

				for (const char* const* c = channels; *c != NULL; c++) {
					this->channels.push_back(std::string(*c));
				}
			}

		public:
			/**
			@param host The remote host to connect to
			@param port The remote port to connect to
			@param disconnectSlow Whether we should be disconnected if we don't keep up.
			@param connectMode the connection mode to use. Set to CONNECT_LEGACY for single connection, all events delivered to
							the default channel. Set to CONNECT_PARALLEL for connection per channel and channel values passed through.
			@param channels An array of names representing the channels to subscribe to. This is a null-terminated array
							of pointers to zero-terminated char arrays. If this is null or empty, subscribe to all channels.
							Note that these channel names should be encoded in UTF-8.
			*/
			EngineConnection(const char * host, int port, bool disconnectSlow, EngineManagement::ConnectMode connectMode, 
											const char * const * channels) {
				init(host, port, disconnectSlow, connectMode, channels);
			}

			/**
			@param host The remote host to connect to
			@param port The remote port to connect to
			@param channels An array of names representing the channels to subscribe to. This is a null-terminated array
							of pointers to zero-terminated char arrays. If this is null or empty, subscribe to all channels.
							Note that these channel names should be encoded in UTF-8.
			*/
			EngineConnection(const char * host, int port, const char * const * channels) {
				init(host, port, false, EngineManagement::CONNECT_LEGACY, channels);
			}

			/**
			Get the channels associated with this connection.
			@return The list of channels that the downstream component should listen
			to.
			*/
			virtual const std::vector<std::string>& getChannels() const {
				return channels;
			}

			/**
			Get the host associated with this connection.
			@return The host of the upstream (even-sending) component
			*/
			virtual const std::string& getHost() const {
				return host;
			}

			/**
			Get the port associated with this connection.
			@return The port of the upstream (even-sending) component
			*/
			virtual int getPort() const {
				return port;
			}

			/**
			Get the value of disconnectSlow flag associated with this connection.
			@return returns true if we should be disconnected if this connection is slow.
			*/
			virtual bool getDisconnectSlow() const{
				return disconnectSlow;
			}

			/**
			Get the connection mode associated with this connection.
			@return The connection mode associated with this connection.
			*/
			virtual EngineManagement::ConnectMode getConnectMode() const {
				return connectMode;
			}

			/**	Stream output operator for a reference. */
			inline friend std::ostream& operator << (std::ostream& stream, const EngineConnection& obj) {
				std::ostringstream ost;
				ost << "EngineConnection to " << obj.host << ":" << obj.port << (obj.disconnectSlow == true ? " diconnectSlow" : "") << " on" <<
					(obj.connectMode == EngineManagement::CONNECT_PARALLEL ? " parallel" : "") << " channels [";

				if (!obj.channels.empty()) {
					std::copy(obj.channels.begin(), obj.channels.end() - 1, std::ostream_iterator<std::string>(ost, ", "));
					ost << obj.channels.back();
				}
				ost << "]";
				stream << ost.str();
				return stream;
			}

			/**	Stream output operator for a pointer. */
			inline friend std::ostream& operator << (std::ostream& stream, const EngineConnection* obj) {
				stream << *obj;
				return stream;
			}

			/** Destroy the underlying connection data. */
			virtual ~EngineConnection() {
			}
		};

		/**
		 * This function must be called once per process first before
		 * any other Engine operations are carried out.
		 *
 		 * @exception EngineException
		 */
		AP_ENGINE_CLIENT_API void engineInit(const char* processName = "C++ Client");


		/**
		 * This function (or engineInit) must be called once per process first before
		 * any other Engine operations are carried out.
		 *
 		 * @exception EngineException
		 */
		AP_ENGINE_CLIENT_API void engineInitMessaging(const char *processName, bool initMessaging=true);

		/**
		 * This function must be called once per process before the
		 * application closes down. It ensures that communications are
		 * shutdown properly and state cleaned up.
		 */
		AP_ENGINE_CLIENT_API void engineShutdown();


		/**
		 * This function attempts to establish a connection to an
		 * Engine. It returns an EngineManagement object that can be used
		 * to access the Engine.
		 *
		 * @param host The machine name where an Engine can be located.
		 * @param port The port that the Engine is listening on for
		 * transport negotiations with the client.
		 * @param disconnectSlowReceiver Disconnect any receiver if it is slow, defaults to false
		 * @param processName if not null, the name of this component.
		 * @return An EngineManagement object if an Engine is located and
		 * a connection established.
 		 * @exception EngineException
		 */
		AP_ENGINE_CLIENT_API EngineManagement* connectToEngine(const char* host, unsigned short port, bool disconnectSlowReceiver = false, const char *processName = nullptr);

		/**
		 * Attempt to establish a receive-only connection to an Engine listening
		 * on the named host and port.  If successful, the returned
		 * EngineManagement can be used for the connectEventConsumer(),
		 * getStatus(), inspectEngine() and ping() operations.
		 * All other operations will fail.
		 *
		 * @param host The hostname of the machine the Engine is running on.
		 * @param port The port that the Engine is listening on.
		 * @param processName if not null, the name of this component.
		 * @return An EngineManagement object operating in receive-only mode.
		 * @exception EngineException If the connection cannot be established.
		 */
		AP_ENGINE_CLIENT_API EngineManagement* connectToEngineReceiveOnly(const char* host, unsigned short port, const char *processName = nullptr);

		/**
		 * Attempt to establish a monitor-only connection to an Engine listening
		 * on the named host and port.  If successful, the returned
		 * EngineManagement can be used for the getStatus(),
		 * inspectEngine() and ping() operations.  All other operations will fail.
		 *
		 * @param host The hostname of the machine the Engine is running on.
		 * @param port The port that the Engine is listening on.
		 * @param processName if not null, the name of this component.
		 * @return An EngineManagement object operating in monitor-only mode.
		 * @exception EngineException If the connection cannot be established.
		 */
		AP_ENGINE_CLIENT_API EngineManagement* connectToEngineMonitorOnly(const char* host, unsigned short port, const char *processName = nullptr);

		/**
		 * This function allows disconnection from an Engine.
		 *
		 * Any EventSuppliers created from the EngineManagement must be
		 * deleted explicitly using deleteEventSupplier in addition to
		 * disconnecting the EngineManagement object.
		 *
		 * @param corr	The Engine to disconnect from.
		 */
		AP_ENGINE_CLIENT_API void disconnectFromEngine(EngineManagement* corr);

		/**
		 * This function allows creation of MonitorScript objects.
		 *
		 * @param monitorString	EPL monitors or event definitions
		 * @return A MonitorScript object.
		 */
		AP_ENGINE_CLIENT_API MonitorScript* createMonitorScript(const char* monitorString);

		/**
		 * This function allows deletion of MonitorScript objects.
		 *
		 * @param mon	The MonitorScript object to delete.
		 */
		AP_ENGINE_CLIENT_API void deleteMonitorScript(MonitorScript* mon);

		/**
		 * This function allows deletion of an EngineInfo object.
		 * All objects returned by any calls to the methods of the
		 * EngineInfo object are also deleted, so after calling
		 * inspectEngine, deleteInfo is the only method which
		 * needs to be called to clean up.
		 *
		 * @param info	The EngineInfo object to delete.
		 */
		AP_ENGINE_CLIENT_API void deleteInfo(EngineInfo* info);


		/**
		 * This function allows deletion of the lists of warnings
		 * returned by injectMonitorScriptWithWarnings(), injectCDPWithWarnings()
		 * and injectJavaWithWarnings()
		 *
		 * @param warnings A list of warnings to be deleted
		 */
		AP_ENGINE_CLIENT_API void deleteWarnings(const char* const* warnings);

		/**
		 * Set custom parameters for this instance of the client library.
		 * Must be called before engineInit. The parameters are passed
		 * as a set of name, value pairs formatted as
		 \verbatim<name1>=<value1>;<name2=value2>;\endverbatim
		 * (i.e. equals sign between name and value, and semicolon separating
		 * the name value pairs).
		 *
		 * The following parameters are defined so far.
		 *
		 * ConfigFile - A filename from which to read a configuration file
		 */
		AP_ENGINE_CLIENT_API void setEngineParams(const char* params);

	} // namespace engine
		
} // namespace apama

} // namespace com


#endif // ENGINE_CLIENT_CPP_HPP
