
/*
 * Title:        plugin_logging.hpp
 * Description:  C++ API for logging from plugins
 * $Copyright (c) 2015-2017, 2019-2020, 2022 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 */

/** @file
 *
 * Contains the headers needed to implement your own Connectivity Plugins.
 */

#ifndef _SAG_PLUGIN_LOGGING_HPP_
#define _SAG_PLUGIN_LOGGING_HPP_

#include <string>

namespace com {
namespace softwareag {
namespace _DATAT_INTERNAL_CPP_NAMESPACE { 

/**
 * \internal
 * Class for enabling per-thread cache optimizations on threads created by transports.
 * This should be created and held live for the whole life time of one simultaneous thread.
 * If you have multiple threads executing simultaneously then you should have multiple CacheOwners.
 */
class CacheOwner
{
	friend class CacheForThisThread;
public:
	/** Create a CacheOwner */
	CacheOwner();
	/** Destroy a CacheOwner */
	~CacheOwner();
private:
	/// \internal
	void *cache;
	/// \internal
	bool inuse;
};

/**
 * \internal
 * Class for enabling per-thread cache optimizations on threads created by transports.
 * This should be created on the stack for the duration of a method call on the thread.
 * Uses the cache owned by the provided CacheOwner.
 * You will get better performance by using the same CacheOwner in the thread that things
 * are allocated as when they are destroyed.
 */
class CacheForThisThread
{
public:
	/** Enable caching using this CacheOwner on this thread */
	CacheForThisThread(CacheOwner &owner);
	/** Disables caching on this thread */
	~CacheForThisThread();
private:
	CacheOwner &owner;
};


/**
 * Class for writing to the system logger.
 *
 * Log messages will be prefixed with the category.
 *
 * Messages will appear for a certain log level if the system log level is at least as verbose.
 */
class Logger
{
public:
	/** The available log levels */
	typedef sag_log_level_t level_t;
	
	/** Construct a logger for the given category */
	Logger(const std::string &category): category(category), prefix(category), cachedLevel(getLevelImpl()) {}
	/**
	 * Log a message at a given level. Templated over the log level.
	 *
	 * Should be called as:
	 *
	 * log<SAG_LOG_CRIT>("format", args);
	 */
	template<level_t LEVEL, typename... ARGS>
	void log(const char *format, ARGS... args) const;

	/**
	 * Log a message at a given level. Level specified at runtime
	 *
	 * Should be called as:
	 *
	 * log(SAG_LOG_CRIT, "format", args);
	 */
	template<typename... ARGS>
	void log(level_t LEVEL, const char *format, ARGS... args) const;
	/**
	 * Log a message at a given level. Templated over the log level.
	 *
	 * Should be called as:
	 *
	 * log<SAG_LOG_CRIT>("format", args);
	 */
	template<level_t LEVEL, typename... ARGS>
	void log(const std::string &format, ARGS... args) const;

	/**
	 * Log a message at a given level. Level specified at runtime
	 *
	 * Should be called as:
	 *
	 * log(SAG_LOG_CRIT, "format", args);
	 */
	template<typename... ARGS>
	void log(level_t LEVEL, const std::string &format, ARGS... args) const;


	/** Returns true if messages at the given level will be written to the log file */
	bool isEnabled(level_t LEVEL) const
	{
		return LEVEL <= getLevel();
	}

	/** Returns the current log level */
	level_t getLevel() const {
		return cachedLevel;
	}

	/** Returns true if CRIT messages will be written to the log */
	bool isCritEnabled() const { return isEnabled(SAG_LOG_CRIT); }
	/** Returns true if FATAL messages will be written to the log */
	bool isFatalEnabled() const { return isEnabled(SAG_LOG_FATAL); }
	/** Returns true if ERROR messages will be written to the log */
	bool isErrorEnabled() const { return isEnabled(SAG_LOG_ERROR); }
	/** Returns true if WARN messages will be written to the log */
	bool isWarnEnabled() const { return isEnabled(SAG_LOG_WARN); }
	/** Returns true if INFO messages will be written to the log */
	bool isInfoEnabled() const { return isEnabled(SAG_LOG_INFO); }
	/** Returns true if DEBUG messages will be written to the log */
	bool isDebugEnabled() const { return isEnabled(SAG_LOG_DEBUG); }
	/** Returns true if TRACE messages will be written to the log */
	bool isTraceEnabled() const { return isEnabled(SAG_LOG_TRACE); }

	/** Log a message at CRIT level */
	template<typename... ARGS> void crit(const char *format, ARGS... args) const { log<SAG_LOG_CRIT>(format, args...); }
	/** Log a message at FATAL level */
	template<typename... ARGS> void fatal(const char *format, ARGS... args) const { log<SAG_LOG_FATAL>(format, args...); }
	/** Log a message at ERROR level */
	template<typename... ARGS> void error(const char *format, ARGS... args) const { log<SAG_LOG_ERROR>(format, args...); }
	/** Log a message at WARN level */
	template<typename... ARGS> void warn(const char *format, ARGS... args) const { log<SAG_LOG_WARN>(format, args...); }
	/** Log a message at INFO level */
	template<typename... ARGS> void info(const char *format, ARGS... args) const { log<SAG_LOG_INFO>(format, args...); }
	/** Log a message at DEBUG level */
	template<typename... ARGS> void debug(const char *format, ARGS... args) const { log<SAG_LOG_DEBUG>(format, args...); }
	/** Log a message at TRACE level */
	template<typename... ARGS> void trace(const char *format, ARGS... args) const { log<SAG_LOG_TRACE>(format, args...); }

	/** Log a message at CRIT level without any arguments */
	void crit(const char *str) const { log<SAG_LOG_CRIT>("%s", str); }
	/** Log a message at FATAL level without any arguments */
	void fatal(const char *str) const { log<SAG_LOG_FATAL>("%s", str); }
	/** Log a message at ERROR level without any arguments */
	void error(const char *str) const { log<SAG_LOG_ERROR>("%s", str); }
	/** Log a message at WARN level without any arguments */
	void warn(const char *str) const { log<SAG_LOG_WARN>("%s", str); }
	/** Log a message at INFO level without any arguments */
	void info(const char *str) const { log<SAG_LOG_INFO>("%s", str); }
	/** Log a message at DEBUG level without any arguments */
	void debug(const char *str) const { log<SAG_LOG_DEBUG>("%s", str); }
	/** Log a message at TRACE level without any arguments */
	void trace(const char *str) const { log<SAG_LOG_TRACE>("%s", str); }


	/** Log a message at CRIT level without any arguments */
	void crit(const std::string &str) const { log<SAG_LOG_CRIT>("%s", str.c_str()); }
	/** Log a message at FATAL level without any arguments */
	void fatal(const std::string &str) const { log<SAG_LOG_FATAL>("%s", str.c_str()); }
	/** Log a message at ERROR level without any arguments */
	void error(const std::string &str) const { log<SAG_LOG_ERROR>("%s", str.c_str()); }
	/** Log a message at WARN level without any arguments */
	void warn(const std::string &str) const { log<SAG_LOG_WARN>("%s", str.c_str()); }
	/** Log a message at INFO level without any arguments */
	void info(const std::string &str) const { log<SAG_LOG_INFO>("%s", str.c_str()); }
	/** Log a message at DEBUG level without any arguments */
	void debug(const std::string &str) const { log<SAG_LOG_DEBUG>("%s", str.c_str()); }
	/** Log a message at TRACE level without any arguments */
	void trace(const std::string &str) const { log<SAG_LOG_TRACE>("%s", str.c_str()); }

	/** Log a message at CRIT level */
	template<typename... ARGS> void crit(const std::string &format, ARGS... args) const { log<SAG_LOG_CRIT>(format, args...); }
	/** Log a message at FATAL level */
	template<typename... ARGS> void fatal(const std::string &format, ARGS... args) const { log<SAG_LOG_FATAL>(format, args...); }
	/** Log a message at ERROR level */
	template<typename... ARGS> void error(const std::string &format, ARGS... args) const { log<SAG_LOG_ERROR>(format, args...); }
	/** Log a message at WARN level */
	template<typename... ARGS> void warn(const std::string &format, ARGS... args) const { log<SAG_LOG_WARN>(format, args...); }
	/** Log a message at INFO level */
	template<typename... ARGS> void info(const std::string &format, ARGS... args) const { log<SAG_LOG_INFO>(format, args...); }
	/** Log a message at DEBUG level */
	template<typename... ARGS> void debug(const std::string &format, ARGS... args) const { log<SAG_LOG_DEBUG>(format, args...); }
	/** Log a message at TRACE level */
	template<typename... ARGS> void trace(const std::string &format, ARGS... args) const { log<SAG_LOG_TRACE>(format, args...); }
	const std::string &getPrefix() const { return prefix; }

private:
	level_t getLevelImpl() const;
	std::string category;
	std::string prefix;
	level_t cachedLevel;
};

}

namespace connectivity { using namespace _DATAT_INTERNAL_CPP_NAMESPACE; }

}} // com.softwareag.connectivity

// internal implementation included from these files
#include <sag_internal/logging.hpp>

#endif // _SAG_PLUGIN_LOGGING_HPP_
